/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

using namespace std;
using namespace FD;

class CvfdhCalcHist;
   
DECLARE_NODE(CvfdhCalcHist)
/*Node
 *
 * @name CvfdhCalcHist
 * @category FDOPENCV_HARK:Hist
 * @description CvfdhCalcHist
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 END*/
   
class CvfdhCalcHist : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_featOutID;
      
public:
  CvfdhCalcHist(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_featOutID = addOutput("FEATOUT");
    inOrder = true;
  }      

  virtual int translateInput(const std::string &inputName) {
    if (inputName == "SRCIN") {
      m_srcInID = addInput(inputName);
      return m_srcInID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;

    RCPtr<Map<int, ObjectRef> > image_map(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      CvfdImage* image= new CvfdImage(img_in_tmp);  
      (*image_map)[0] = image;
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	CvfdImage* image= new CvfdImage(img_in_tmp);  
	(*image_map)[it->first] = image;
      }
    }

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map->begin();
    for (; img_it != image_map->end(); ++img_it) {
      IplImage* img_in;      
      CvfdImage& imagePtr = object_cast<CvfdImage>(img_it->second);
      img_in = cvCloneImage(imagePtr.getImage());
    
      RCPtr<Vector<float> > map_contents(new Vector<float>);
    
      // ---------------------------------------------------------------

      int i, j, bin_w;
      int hist_size = 256;
      int sch = 0, ch_width = 260;
      float max_value = 0;
      float range_0[] = { 0, 256 };
      float *ranges[] = { range_0 };
      IplImage *dst_img[4] = { 0, 0, 0, 0 };
      // IplImage *hist_img;
      CvHistogram *hist;

      sch = img_in->nChannels;
      for (i = 0; i < sch; i++) {
	dst_img[i] = cvCreateImage (cvSize (img_in->width, img_in->height), img_in->depth, 1);
      }
    
      hist = cvCreateHist (1, &hist_size, CV_HIST_ARRAY, ranges, 1);
      // hist_img = cvCreateImage (cvSize (ch_width * sch, 200), 8, 1);
    
      if (sch == 1)
	cvCopy (img_in, dst_img[0], NULL);
      else
	cvSplit (img_in, dst_img[0], dst_img[1], dst_img[2], dst_img[3]);
    
      // cvSet (hist_img, cvScalarAll (255), 0);
      for (i = 0; i < sch; i++) {
	// CalcHist and Nomalization
	cvCalcHist (&dst_img[i], hist, 0, NULL);
	cvGetMinMaxHistValue (hist, 0, &max_value, 0, 0);
	// cvScale (hist->bins, hist->bins, ((double) hist_img->height) / max_value, 0);
	bin_w = cvRound ((double) ch_width / hist_size);
	// Make an image
	for (j = 0; j < hist_size; j++){
	  // cvRectangle (hist_img,cvPoint (j * bin_w + (i * ch_width), hist_img->height), cvPoint ((j + 1) * bin_w + (i * ch_width), hist_img->height - cvRound (cvGetReal1D (hist->bins, j))), cvScalarAll (0), -1, 8, 0);
	  (*map_contents).push_back((float)cvGetReal1D (hist->bins, j));
	}
      }
 
      (*outputMapPtr)[img_it->first] = map_contents;    

      // Output Setting
      //img_in = cvCloneImage (hist_img);
      // End of the transform code            
      // main routine : finish
      // ---------------------------------------------------------------
      // CvfdImage* image= new CvfdImage(img_in);           
      // (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage(&img_in);
      // cvReleaseImage(&hist_img);
      for(int i = 0; i < 4; i++) cvReleaseImage(&dst_img[i]);

    }

    cvReleaseImage(&img_in_tmp);
    
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhCalcHist)
};
