/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhKMeans;
   
DECLARE_NODE(CvfdhKMeans)
/*Node
 *
 * @name CvfdhKMeans
 * @category FDOPENCV_HARK:Filter
 * @description CvfdhKMeans
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @parameter_name KMEANS_MAX_CLUSTERS
 * @parameter_type int
 * @parameter_value 10
 * @parameter_description
 *
 END*/
   
   
class CvfdhKMeans : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
  int MAX_CLUSTERS;
      
public:
  CvfdhKMeans(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    MAX_CLUSTERS = dereference_cast<int>(parameters.get("KMEANS_MAX_CLUSTERS")) ;
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    ObjectRef in = getInput(m_imageInID, count);

    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {

      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
	return;
      }

      IplImage* img_in;
      img_in = cvCloneImage(imagePtr->getImage());
    
      int i, size;
      IplImage *src_img = 0, *dst_img = 0;
      CvMat tmp_header;
      CvMat *clusters, *points, *tmp;
      CvMat *countMat = cvCreateMat( MAX_CLUSTERS, 1, CV_32SC1);
      CvMat *centers = cvCreateMat( MAX_CLUSTERS, 3, CV_32FC1);
      char *imagename;

      // (1)load a specified file as a 3-channel color image
      src_img = cvCloneImage (img_in);

      size = src_img->width * src_img->height;
      dst_img  = cvCloneImage(src_img);
      clusters = cvCreateMat(size, 1, CV_32SC1 );
      points   = cvCreateMat( size, 1, CV_32FC3 );

      // (2)reshape the image to be a 1 column matrix 
#if 0
      tmp = cvCreateMat(size, 1, CV_8UC3);
      tmp = cvReshape(src_img, &tmp_header, 0, size);
      cvConvert(tmp, points);
      cvReleaseMat(&tmp);
#else
      for(i=0; i<size; i++) {
	points->data.fl[i*3+0] = (uchar)src_img->imageData[i*3+0];
	points->data.fl[i*3+1] = (uchar)src_img->imageData[i*3+1];
	points->data.fl[i*3+2] = (uchar)src_img->imageData[i*3+2];
      }
#endif
  
      // (3)run k-means clustering algorithm to segment pixels in RGB color space
#if ENABLE_OPENCV_2_0_0

      cvKMeans2( points, MAX_CLUSTERS, clusters,
		 cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 10, 1.0 ),
		 1, 0, 0, centers, 0);

#else

      cvKMeans2( points, MAX_CLUSTERS, clusters,
		 cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 10, 1.0 ));

#endif

      // (4)make a each centroid represent all pixels in the cluster
      for(i=0; i<size; i++) {
	int idx = clusters->data.i[i];
	dst_img->imageData[i*3+0] = (char)centers->data.fl[idx*3+0];
	dst_img->imageData[i*3+1] = (char)centers->data.fl[idx*3+1];
	dst_img->imageData[i*3+2] = (char)centers->data.fl[idx*3+2];
      }

      IplImage* img_out;
      img_out = cvCloneImage(dst_img);
         
      CvfdImage* image= new CvfdImage(img_out);  
         
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage(&img_in);
      cvReleaseImage(&src_img);
      cvReleaseImage(&dst_img);
      cvReleaseImage(&img_out);


      // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* img_in;
	img_in = cvCloneImage(imagePtr.getImage());

	int i, size;
	IplImage *src_img = 0, *dst_img = 0;
	CvMat tmp_header;
	CvMat *clusters, *points, *tmp;
	CvMat *countMat = cvCreateMat( MAX_CLUSTERS, 1, CV_32SC1);
	CvMat *centers = cvCreateMat( MAX_CLUSTERS, 3, CV_32FC1);
	char *imagename;

	// (1)load a specified file as a 3-channel color image
	src_img = cvCloneImage (img_in);

	size = src_img->width * src_img->height;
	dst_img  = cvCloneImage(src_img);
	clusters = cvCreateMat(size, 1, CV_32SC1 );
	points   = cvCreateMat( size, 1, CV_32FC3 );

	// (2)reshape the image to be a 1 column matrix 
#if 0
	tmp = cvCreateMat(size, 1, CV_8UC3);
	tmp = cvReshape(src_img, &tmp_header, 0, size);
	cvConvert(tmp, points);
	cvReleaseMat(&tmp);
#else
	for(i=0; i<size; i++) {
	  points->data.fl[i*3+0] = (uchar)src_img->imageData[i*3+0];
	  points->data.fl[i*3+1] = (uchar)src_img->imageData[i*3+1];
	  points->data.fl[i*3+2] = (uchar)src_img->imageData[i*3+2];
	}
#endif
  
	// (3)run k-means clustering algorithm to segment pixels in RGB color space

#if ENABLE_OPENCV_2_0_0

	cvKMeans2( points, MAX_CLUSTERS, clusters,
		   cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 10, 1.0 ),
		   1, 0, 0, centers, 0);

#else

	cvKMeans2( points, MAX_CLUSTERS, clusters,
		   cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 10, 1.0 ));
	
#endif

	// (4)make a each centroid represent all pixels in the cluster
	for(i=0; i<size; i++) {
	  int idx = clusters->data.i[i];
	  dst_img->imageData[i*3+0] = (char)centers->data.fl[idx*3+0];
	  dst_img->imageData[i*3+1] = (char)centers->data.fl[idx*3+1];
	  dst_img->imageData[i*3+2] = (char)centers->data.fl[idx*3+2];
	}

	IplImage* img_out;
	img_out = cvCloneImage(dst_img);
         
	CvfdImage* image= new CvfdImage(img_out);  
	(*outputMapPtr)[it->first] = image;
         

	cvReleaseImage(&img_in);
	cvReleaseImage(&src_img);
	cvReleaseImage(&dst_img);
	cvReleaseImage(&img_out);

      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    }

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhKMeans)
};
