/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include <ctype.h>
#include <stdio.h>
#include <vector>

using namespace std;
using namespace FD;

class CvfdhRunningAvg;
   
DECLARE_NODE(CvfdhRunningAvg)
/*Node
 *
 * @name CvfdhRunningAvg
 * @category FDOPENCV_HARK:Filter
 * @description CvfdhRunningAvg
 *
 * @input_name IMAGEIN
 * @input_description Input image
 * @input_type CvfdImage
 *
 * @output_name IMAGEOUT
 * @output_description Output image
 * @output_type CvfdImage
 *
 * @parameter_name WINDOW_SIZE
 * @parameter_type int
 * @parameter_value 20
 * @parameter_description 
 *
 * @parameter_name BGN_B_PARAM
 * @parameter_type float
 * @parameter_value 0.02
 * @parameter_description 
 *
 * @parameter_name BGN_T_PARAM
 * @parameter_type float
 * @parameter_value 0.005
 * @parameter_description 
 *
 * @parameter_name BGN_ZETA
 * @parameter_type float
 * @parameter_value 10.0
 * @parameter_description 
 *
 END*/
   
   
class CvfdhRunningAvg : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
  int BG_WINDOW_SIZE;
  float B_PARAM;
  float T_PARAM;
  float Zeta;
  IplImage* img_out;

  // フレーム間画像格納のためのベクトル
  std::vector<IplImage *> imageVector;
      
public:
  CvfdhRunningAvg(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    BG_WINDOW_SIZE = dereference_cast<int>(parameters.get("WINDOW_SIZE"));
    B_PARAM = dereference_cast<float>(parameters.get("BGN_B_PARAM"));
    T_PARAM = dereference_cast<float>(parameters.get("BGN_T_PARAM"));
    Zeta    = dereference_cast<float>(parameters.get("BGN_ZETA"));
    inputsCache[m_imageInID].lookAhead = BG_WINDOW_SIZE - 1;
    imageVector.resize(0);
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    IplImage *lower_img, *upper_img, *tmp_img;
    IplImage *dst_img, *msk_img;
    IplImage *av_img, *sgm_img;

    int width;
    int height;

    // ---------------------------------------------------------------
    // main routine : start

    CvfdImage* imageOutTmp= new CvfdImage(img_out);  
    (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imageOutTmp); 


    RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
    if(imagePtr->getImage() == NULL){
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
      return;
    }

    IplImage* img_in;
    img_in = cvCloneImage(imagePtr->getImage());

    imageVector.push_back(img_in);
    
    if(imageVector.size() < BG_WINDOW_SIZE) return;

    width = img_in->width;
    height = img_in->height;      
    
    tmp_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_32F, 3);
    lower_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_32F, 3);
    upper_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_32F, 3);
    dst_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_8U, 3);
    msk_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_8U, 1);

    // Accumulate for back ground ---------------------------------------------------------

    // (3)作業用の領域を生成する
    av_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_32F, 3);
    sgm_img = cvCreateImage (cvSize (width, height), IPL_DEPTH_32F, 3);    

    // (4)背景の輝度平均の初期値を計算する(av_imgの作成)
    cvSetZero (av_img);
    for (int i = 0; i < BG_WINDOW_SIZE; i++) {
      cvAcc (imageVector[i], av_img); // av_img = av_img + frame
    }
    cvConvertScale (av_img, av_img, 1.0 / BG_WINDOW_SIZE); // Normalization

    // (5)背景の輝度振幅の初期値を計算する(sgm_imgの作成)
    cvSetZero (sgm_img);
    for (int i = 0; i < BG_WINDOW_SIZE; i++) {
      cvConvert (imageVector[i], tmp_img);
      cvSub (tmp_img, av_img, tmp_img);
      cvPow (tmp_img, tmp_img, 2.0);
      cvConvertScale (tmp_img, tmp_img, 2.0);
      cvPow (tmp_img, tmp_img, 0.5);
      cvAcc (tmp_img, sgm_img);
    }
    cvConvertScale (sgm_img, sgm_img, 1.0 / BG_WINDOW_SIZE); // Normalization

    // main routine ----------------------------------------------------------------------
    
    // (7)取得画像から背景を分離するループ(tmp_img = 現在のフレーム)
    cvConvert (img_in, tmp_img);
    
    // (8)背景となりうる画素の輝度値の範囲をチェックする(lower_img, upper_img, msk_imgの更新)
    cvSub (av_img, sgm_img, lower_img);
    cvSubS (lower_img, cvScalarAll (Zeta), lower_img);
    cvAdd (av_img, sgm_img, upper_img);
    cvAddS (upper_img, cvScalarAll (Zeta), upper_img);
    cvInRange (tmp_img, lower_img, upper_img, msk_img);
  
    // (9)輝度振幅を再計算する(tmp_img = 輝度振幅)
    cvSub (tmp_img, av_img, tmp_img);
    cvPow (tmp_img, tmp_img, 2.0);
    cvConvertScale (tmp_img, tmp_img, 2.0);
    cvPow (tmp_img, tmp_img, 0.5);
  
    // (10)背景と判断された領域の背景の輝度平均と輝度振幅を更新する
    cvRunningAvg (img_in, av_img, B_PARAM, msk_img);    // av_imgの更新
    cvRunningAvg (tmp_img, sgm_img, B_PARAM, msk_img); // sgm_imgの更新
  
    // (11)物体領域と判断された領域では輝度振幅のみを（背景領域よりも遅い速度で）更新する
    cvNot (msk_img, msk_img);                          // 反転(物体領域のみの演算にする)
    cvRunningAvg (tmp_img, sgm_img, T_PARAM, msk_img); // sgm_imgの更新
  
    // (12)物体領域のみを出力画像にコピーする（背景領域は黒）
    cvSetZero (dst_img);
    cvCopy (img_in, dst_img, msk_img);
    
    // (13)処理結果を表示する
    // char str[64];
    // snprintf (str, 64, "%03d[frame]", counter);
    // cvPutText (dst_img, str, cvPoint (10, 20), &font, CV_RGB (0, 255, 100));

    img_out = cvCloneImage(dst_img);

    // main routine : finish
    // ---------------------------------------------------------------
         
    CvfdImage* image= new CvfdImage(img_out);  
    (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

    // Vector reset
    imageVector.resize(0);

    cvReleaseImage (&lower_img); 
    cvReleaseImage (&upper_img);
    cvReleaseImage (&tmp_img);
    cvReleaseImage (&dst_img);
    cvReleaseImage (&msk_img);
    cvReleaseImage (&av_img);
    cvReleaseImage (&sgm_img);
    cvReleaseImage (&img_in);

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhRunningAvg)
};
