/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Source.h"
#include "CvfdCameraSpec.h"
#include "Vector.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhSrcInfoFormatConverter;
   
DECLARE_NODE(CvfdhSrcInfoFormatConverter)
/*Node
 *
 * @name CvfdhSrcInfoFormatConverter
 * @category FDOPENCV_HARK:ROI
 * @description CvfdhSrcInfoFormatConverter
 *
 * @input_name CAMSPEC
 * @input_type CvfdCameraSpec
 * @input_description Camera specification
 *
 * @input_name SRCINFO
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @input_name MAPIN
 * @input_type Map<int,ObjectRef>
 * @input_description Input Features. This is an optional input. The ObjectRef in the Map is Vector<float>.
 *
 * @output_name SOURCE
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is Source.
 *
 * @output_name MAPOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 END*/
   
   
class CvfdhSrcInfoFormatConverter : public BufferedNode {

  int m_camspecID;
  int m_srcinfoID;
  int m_mapinID;
  int m_sourceID;
  int m_mapoutID;
      
public:
  CvfdhSrcInfoFormatConverter(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_mapinID(-1)
  {
    m_camspecID = addInput("CAMSPEC");
    m_srcinfoID = addInput("SRCINFO");
    m_sourceID = addOutput("SOURCE");
    m_mapoutID = addOutput("MAPOUT");
    inOrder = true;
  }      

  virtual int translateInput(const std::string &inputName) {
    if (inputName == "MAPIN") {
      m_mapinID = addInput(inputName);
      return m_mapinID;
    }
    else if ((inputName == "SRCINFO")||(inputName == "CAMSPEC")) {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<CvfdCameraSpec> cspec  = getInput(m_camspecID, count);

    RCPtr<Vector<ObjectRef> > sources  = getInput(m_srcinfoID, count);

    RCPtr<Map<int, ObjectRef> > input;
    if (m_mapinID != -1) {
      input = getInput(m_mapinID, count);
    }

    Vector<ObjectRef>& srcout = *new Vector<ObjectRef>;
    (*(outputs[m_sourceID].buffer))[count] = &srcout;

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_mapoutID].buffer))[count] = outputMapPtr;

    for (int i = 0; i < sources->size(); i++) {

      RCPtr<SrcInfo> srci = (*sources)[i];

      float theta = (*cspec).pt2rad_x(srci->roi_c[0]);
      float phi   = (*cspec).pt2rad_y(srci->roi_c[1]);
      float cos_phi = cos(phi);

      RCPtr<Source> srco(new Source);
      srco->id  = srci->id;
      srco->power = 0.0;
      srco->x[0] = cos(theta) * cos_phi;
      srco->x[1] = sin(theta) * cos_phi;
      srco->x[2] = sin(phi);
      srcout.push_back(srco);

      RCPtr<Vector<float> > map_contents(new Vector<float>);
      (*map_contents).push_back((float)srci->roi_c[0]);
      (*map_contents).push_back((float)srci->roi_c[1]);
      (*map_contents).push_back((float)srci->roi_xy[0]);
      (*map_contents).push_back((float)srci->roi_xy[1]);

      if (m_mapinID != -1) {
	Map<int, ObjectRef>::iterator it = input->find(srci->id);
	if (it != input->end()) {
	  const Vector<float>& feature = object_cast<Vector<float> >(it->second);
	  for(int cnt = 0; cnt < feature.size(); cnt++) (*map_contents).push_back((float)feature[cnt]);
	}
      }
      (*outputMapPtr)[srci->id] = map_contents;    
    }
  } 
  NO_ORDER_NODE_SPEEDUP(CvfdhSrcInfoFormatConverter)
};
