/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "CvfdColor.h"

using namespace std;
using namespace FD;
   
class CvfdFindColorHSV;
   
DECLARE_NODE(CvfdFindColorHSV)
/*Node
 *
 * @name CvfdFindColorHSV
 * @category FDOPENCV_ORG:Color
 * @description Detects the color in the image with a HSV method
 * dst(I) is set to 0xff (all '1'-bits) if the color is detect and 0 otherwise.
 * 
 * @input_name IMAGEIN
 * @input_description Input image
 * @input_type CvfdImage
 *
 * @input_name COLORMIN
 * @input_description Color which represents the minimal value for each channel.
 * @input_type CvfdColor
 *
 * @input_name COLORMAX
 * @input_description Color which represents the maximal value for each channel.
 * @input_type CvfdColor
 *
 * @output_name IMAGEOUT
 * @output_description Result image
 * @output_type CvfdImage
 *
 END*/
inline std::string CINT(int value)
{
  std::stringstream sstr;
  sstr << value;
  return(sstr.str());
}   
   
class CvfdFindColorHSV : public BufferedNode {
      
  //Input ID
  int m_imageInID;
  int m_colorMinID;
  int m_colorMaxID;
  //Output ID
  int m_imageOutID;
      
public:
  CvfdFindColorHSV(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_imageInID = addInput("IMAGEIN");
    m_colorMinID = addInput("COLORMIN");
    m_colorMaxID = addInput("COLORMAX");
         
    //add outputs
    m_imageOutID = addOutput("IMAGEOUT");
         
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {     
    //Read the inputs         
    RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);
    RCPtr<CvfdColor> colorMinPtr = getInput(m_colorMinID,count);
    RCPtr<CvfdColor> colorMaxPtr = getInput(m_colorMaxID,count);
         
    CvfdImage* image = new CvfdImage( cvGetSize(imagePtr->getImage()), IPL_DEPTH_8U, 1 ); 
    IplImage* imageHSV; 
         
    int status = cvGetErrMode();
    cvSetErrMode( CV_ErrModeSilent );
    __BEGIN__;     
    OPENCV_CALL( imageHSV = cvCreateImage( cvGetSize(imagePtr->getImage()), IPL_DEPTH_8U, 3));
    OPENCV_CALL( cvCvtColor( imagePtr->getImage(), imageHSV, CV_BGR2HSV )); 
         
    __END__;
    cvSetErrMode( status );
    if( cvGetErrStatus() != CV_StsOk  )
      {
	throw new GeneralException("OPENCV - Error to find the color: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      } 
          
    unsigned char H,S,V;
    bool h;
    int x,y;
    for (x=0; x<imageHSV->width ; x++)
      {
	for (y=0; y<imageHSV->height ; y++)
	  {
	    H = ((uchar*)(imageHSV->imageData + imageHSV->widthStep*y))[x*3];
	    S = ((uchar*)(imageHSV->imageData + imageHSV->widthStep*y))[x*3+1];
	    V = ((uchar*)(imageHSV->imageData + imageHSV->widthStep*y))[x*3+2];

	    if ( colorMaxPtr->channel1() > colorMinPtr->channel1())
	      {
		h = ( H*2 <= colorMaxPtr->channel1() ) && ( H*2 >= colorMinPtr->channel1() );
	      }
	    else
	      {
		h = ( H*2 <= colorMaxPtr->channel1() ) || ( H*2 >= colorMinPtr->channel1() );
	      }
               
	    if( h
		&& ( S <= colorMaxPtr->channel2()) && ( S >= colorMinPtr->channel2() )
		&& ( V <= colorMaxPtr->channel3() ) && ( V >= colorMinPtr->channel3() ) )
               
	      {
		((uchar*)(image->getImage()->imageData + image->getImage()->widthStep*y))[x]=255;
	      }               
	  }
      } 
    cvReleaseImage( &imageHSV );
    out[count] = ObjectRef(image);
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdFindColorHSV)
};
